// // >nul 2> nul & @goto :main
/*
 * Author: penpen
 * Extended: aGerman
 * Dependencies: XP upwards, .NET Framework 4 or higher
 * Free for non profit use only.
 * For profit use contact me at www.dostips.com via private message (PM).
 * CmdFont.cs.bat
 */

/*
:main
    @echo off
    setlocal
    cls

    set "csc="

    pushd "%SystemRoot%\Microsoft.NET\Framework"
    for /f "tokens=* delims=" %%i in ('dir /b /o:n "v*"') do (
        dir /a-d /b "%%~fi\csc.exe" >nul 2>&1 && set "csc="%%~fi\csc.exe""
    )
    popd

    if defined csc (
        echo most recent C#.NET compiler located in:
        echo %csc%.
    ) else (
        echo C#.NET compiler not found.
        goto :eof
    )

    for %%a in ("%~dpn0") do for %%b in ("%%~dpna") do (
        %csc% /nologo /optimize /warnaserror /nowin32manifest /unsafe /debug- /target:exe /out:"%%~b.exe" "%~f0"
    )

    exit /B
*/

using System;
using System.Linq;
using System.Drawing;
using System.Reflection;
using System.Collections.Generic;
using System.Runtime.InteropServices;

// translation table for types used in API functions
using BOOL   = System.Int32;
using BYTE   = System.Byte;
using DWORD  = System.Int32;
using HANDLE = System.IntPtr;
using HDC    = System.IntPtr;
using LONG   = System.Int32;
using LPARAM = System.IntPtr; /* should have been either Int32 or Int64 depending on the target platform
                                 but IntPtr is at least guaranteed to have always the right size */
using SHORT  = System.Int16;
using UINT   = System.Int32;
using ULONG  = System.Int32;
using WCHAR  = System.Char;

// add version information to the "Details" properties
[assembly:AssemblyTitleAttribute("Getting And Setting Console Fonts")]
[assembly:AssemblyVersionAttribute("2.1.0.0")]
[assembly:AssemblyCopyrightAttribute("(c) 2017 penpen (dostips.com)")]
[assembly:AssemblyProductAttribute("CmdFont")]

// class for API constants
public static class Constants {
    public const Int32 FALSE             = 0;
    public const Int32 TRUE              = 1;
    public const Int32 LF_FACESIZE       = 32;
    public const Int32 LF_FULLFACESIZE   = 64;
    public const Int32 OEM_CHARSET       = 0x000000FF;
    public const uint GENERIC_WRITE      = 0x40000000;
    public const uint GENERIC_READ       = 0x80000000;
    public const Int32 FILE_SHARE_WRITE  = 0x00000002;
    public const Int32 OPEN_EXISTING     = 3;
}

// global structures used in the Windows API and the SortedSet of found fonts
[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
public unsafe struct LOGFONTW {
    public LONG lfHeight;
    public LONG lfWidth;
    public LONG lfEscapement;
    public LONG lfOrientation;
    public LONG lfWeight;
    public BYTE lfItalic;
    public BYTE lfUnderline;
    public BYTE lfStrikeOut;
    public BYTE lfCharSet;
    public BYTE lfOutPrecision;
    public BYTE lfClipPrecision;
    public BYTE lfQuality;
    public BYTE lfPitchAndFamily;
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst = Constants.LF_FACESIZE)]
    public String lfFaceName;
}

[StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
public unsafe struct ENUMLOGFONTW {
    public LOGFONTW elfLogFont;
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst = Constants.LF_FULLFACESIZE)]
    public String elfFullName;
    [MarshalAs(UnmanagedType.ByValTStr, SizeConst = Constants.LF_FACESIZE)]
    public String elfStyle;
}

// custom class used for sorting of elements in a SortedSet<ENUMLOGFONTW>
public class EnumLogFontComparer : IComparer<ENUMLOGFONTW> {
    public int Compare (ENUMLOGFONTW x, ENUMLOGFONTW y) {
        int result = x.elfLogFont.lfHeight.CompareTo (y.elfLogFont.lfHeight);
        return result == 0 ? x.elfLogFont.lfWidth.CompareTo (y.elfLogFont.lfWidth) : result;
    }
}

// program class
public class CmdFont {

    // internally used API structures
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    internal unsafe struct NEWTEXTMETRICW {
        internal LONG tmHeight;
        internal LONG tmAscent;
        internal LONG tmDescent;
        internal LONG tmInternalLeading;
        internal LONG tmExternalLeading;
        internal LONG tmAveCharWidth;
        internal LONG tmMaxCharWidth;
        internal LONG tmWeight;
        internal LONG tmOverhang;
        internal LONG tmDigitizedAspectX;
        internal LONG tmDigitizedAspectY;
        internal WCHAR tmFirstChar;
        internal WCHAR tmLastChar;
        internal WCHAR tmDefaultChar;
        internal WCHAR tmBreakChar;
        internal WCHAR tmItalic;
        internal WCHAR tmUnderlined;
        internal WCHAR tmStruckOut;
        internal WCHAR tmPitchAndFamily;
        internal WCHAR tmCharSet;
        internal DWORD ntmFlags;
        internal UINT ntmSizeEM;
        internal UINT ntmCellHeight;
        internal UINT ntmAvgWidth;
    }

    [StructLayout (LayoutKind.Sequential)]
    internal struct COORD {
        internal SHORT X;
        internal SHORT Y;
    }

    [StructLayout (LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    internal unsafe struct CONSOLE_FONT_INFOEX {
        internal ULONG cbSize;
        internal DWORD nFont;
        internal COORD dwFontSize;
        internal UINT FontFamily;
        internal UINT FontWeight;
        [MarshalAs(UnmanagedType.ByValTStr, SizeConst = Constants.LF_FACESIZE)]
        internal String FaceName;
    }

    // delegate used instead of the function pointer for EnumFontFamiliesExW
    private delegate int FONTENUMPROC (ref ENUMLOGFONTW lpelf, ref NEWTEXTMETRICW lpntm, DWORD FontType, LPARAM lParam);
    // sorted set of found raster fonts
    private static SortedSet<ENUMLOGFONTW> fontDataSet;
    // constants for program options
    private const int HELP = 0;
    private const int GET = 1;
    private const int SET = 2;

    // imported API functions
    [DllImport("gdi32.dll", CharSet = CharSet.Unicode, ExactSpelling=true, SetLastError=true)]
    private static extern int EnumFontFamiliesExW (
        HDC                     hdc,
        ref LOGFONTW            lpLogfont,
        FONTENUMPROC            lpEnumFontFamExProc,
        LPARAM                  lParam,
        DWORD                   dwFlags
    );

    [DllImport("kernel32.dll", CharSet=CharSet.Auto, ExactSpelling=true, SetLastError=true)]
    private static extern BOOL SetCurrentConsoleFontEx (
        HANDLE                  consoleOutput,
        BOOL                    maximumWindow,
        ref CONSOLE_FONT_INFOEX consoleCurrentFontEx
    );

    [DllImport ("kernel32.dll", CharSet = CharSet.Unicode, ExactSpelling=true, SetLastError=true)]
    private static extern HANDLE CreateFileW(
        string lpFileName,
        uint dwDesiredAccess,
        DWORD dwShareMode,
        IntPtr SecurityAttributes,
        DWORD dwCreationDisposition,
        DWORD dwFlagsAndAttributes,
        HANDLE hTemplateFile
    );

    [DllImport ("kernel32.dll", CharSet=CharSet.Auto, ExactSpelling=true, SetLastError=true)]
    private static extern BOOL GetCurrentConsoleFontEx (
        HANDLE                  consoleOutput,
        BOOL                    maximumWindow,
        ref CONSOLE_FONT_INFOEX lpConsoleCurrentFont
    );

    [DllImport ("kernel32.dll", CharSet=CharSet.Auto, ExactSpelling=true, SetLastError=true)]
    private static extern COORD GetConsoleFontSize (
        HANDLE                  hConsoleOutput,
        DWORD                   nFont
    );

    // callback method for EnumFontFamiliesExW
    private static unsafe int EnumFontFamExProc (ref ENUMLOGFONTW lpelf, ref NEWTEXTMETRICW lpntm, DWORD FontType, LPARAM lParam) {
        fontDataSet.Add (lpelf);
        return 1;
    }

    // program entry point
    static int Main (string[] args) {
        bool err = false;
        int option = HELP;
        int nFont = 0;

        // process program options
        switch (args.Length) {
            case 0:
                option = GET;
                break;

            case 1:
                option = (String.Compare ("GET", args [0], true) == 0) ? GET : HELP;
                break;

            case 2:
                try {
                    nFont = int.Parse (args [1]);
                    option = (String.Compare ("SET", args [0], true) == 0) ? SET : HELP;
                } catch {
                    err = true;
                }
                break;

            default:
                option = HELP;
                break;
        }

        // enumerate raster fonts
        fontDataSet = new SortedSet<ENUMLOGFONTW> (new EnumLogFontComparer ());
        Graphics graphics = Graphics.FromHwnd (IntPtr.Zero);
        try {
            HDC hdc = graphics.GetHdc ();
            var logfont = new LOGFONTW () { lfCharSet = (BYTE)Constants.OEM_CHARSET, lfFaceName = "Terminal" };
            EnumFontFamiliesExW (hdc, ref logfont, new FONTENUMPROC (EnumFontFamExProc), IntPtr.Zero, 0);
        } catch {
            err = true;
        } finally {
            graphics.ReleaseHdc ();
        }

        if (err || fontDataSet.Count == 0) {
          return 1;
        }

        // get console output handle
        HANDLE hOut = CreateFileW ("CONOUT$", Constants.GENERIC_READ | Constants.GENERIC_WRITE, Constants.FILE_SHARE_WRITE, IntPtr.Zero, Constants.OPEN_EXISTING, 0, IntPtr.Zero);

        // core functionality depending on the passed options
        switch (option) {
            // set new font
            case SET:
                if (fontDataSet.Count <= nFont || nFont < 0) {
                    return 1;
                }

                {
                    ENUMLOGFONTW font = fontDataSet.ElementAt (nFont);
                    CONSOLE_FONT_INFOEX infoex = new CONSOLE_FONT_INFOEX () {
                        cbSize = (ULONG)Marshal.SizeOf(typeof(CONSOLE_FONT_INFOEX)),
                        nFont = nFont,
                        dwFontSize = new COORD {X = (SHORT)font.elfLogFont.lfWidth, Y = (SHORT)font.elfLogFont.lfHeight},
                        FontFamily = 48,
                        FontWeight = 400,
                        FaceName = "Terminal"
                    };
                    if (SetCurrentConsoleFontEx (hOut, Constants.FALSE, ref infoex) == 0) {
                        return 1;
                    }
                }
                break;

            // list fonts
            case GET:
                {
                    CONSOLE_FONT_INFOEX infoex = new CONSOLE_FONT_INFOEX () { cbSize = (ULONG)Marshal.SizeOf(typeof(CONSOLE_FONT_INFOEX)) };
                    if (GetCurrentConsoleFontEx (hOut, Constants.FALSE, ref infoex) == 0) {
                        return 1;
                    }
                    COORD fontSize = GetConsoleFontSize (hOut, infoex.nFont);

                    bool found = false;
                    int i = 0;
                    foreach (ENUMLOGFONTW font in fontDataSet) {
                        Console.Write ("{0}::{1}::{2}x{3}", i++, font.elfFullName, font.elfLogFont.lfWidth, font.elfLogFont.lfHeight);
                        if (!found
                            && font.elfLogFont.lfWidth == fontSize.X
                            && font.elfLogFont.lfHeight == fontSize.Y
                            && font.elfFullName == infoex.FaceName
                        ) {
                            Console.WriteLine ("::CURRENT");
                            found = true;
                        } else {
                            Console.WriteLine ();
                        }
                    }

                    if (!found) {
                        Console.WriteLine ("-::{0}::{1}::CURRENT", infoex.FaceName, fontSize.Y);
                    }

                    Console.Out.Flush ();
                }
                break;

            // help message
            default:
                Console.WriteLine ("Usage CmdFont[.exe] [GET|SET nFont]\r\n" +
                                   "  [GET]      Displays a message with a list of supported fonts.\r\n" +
                                   "             Format <nFont>::<Font Name>::<Font Size>[::CURRENT]\r\n" +
                                   "             If the current font is not supported the line begins\r\n" +
                                   "             with - instead of nFont.\r\n" +
                                   "  SET nFont  Sets the font to \"Terminal\" (Raster Font) with the\r\n" +
                                   "             size that is related to nFont.");
                Console.Out.Flush ();
                break;
        }

       return 0;
    }
}
